package edu.cmu.cs211.graph;
import static edu.cmu.cs211.graph.GraphSearchUtil.*;
import java.util.*;

/**
 * Searches a Graph using Bellman-Ford. This implementation of GraphSearch
 * can handle a graph with negative edge weights
 *
 */
public class BellmanFord implements GraphSearch {
	/**
	 * @throws IllegalArgumentException If <tt>graph</tt> has a negative cost cycle.
	 */
	public <V, E extends Edge<V>> List<E> shortestPath(Graph<V, E> graph, V start, V finish) {
		// Represents a path from the start to some vertex (via another vertex), and the
		// associated cost.
		class Path {
			public V previous; // The vertex prior the destination in the path
			// The cost of the trip; that is, the sum of the weights of the edges
			// that make up the path.
			public int cost;
			public int timesEnqueued = 0;
			public Path() {
				previous = null;
				cost = INFINITY;
			}
			public Path(int cost, V previous) {
				this.cost = cost;
				this.previous = previous;
			}
		}
		// We use a FIFO queue to store a list of vertices that need to be processed.
		LinkedList<V> vertexQueue = new LinkedList<V>();
		// A collection of paths going from the start to every vertex in the path, where
		// the key in the map is the destination vertex.
		Map<V, Path> pathMap = new HashMap<V, Path>();
		// Initialize the map with the vertices in the graph
		for(V vertex : graph.vertices())
			pathMap.put(vertex, new Path());
		pathMap.put(start, new Path(0, start));
		vertexQueue.offer(start);
		pathMap.get(start).timesEnqueued++;
		while(!vertexQueue.isEmpty()) {
			V src = vertexQueue.poll();
			Path srcPath = pathMap.get(src);
			// Relax the neighbors of the vertex
			for(E edge : graph.outgoingEdges(src)) {
				Path destPath = pathMap.get(edge.dest());
				int alternateCost = safeAdd(srcPath.cost, edge.weight());
				if(alternateCost < destPath.cost) {
					Path newPath = new Path(alternateCost, src);
					newPath.timesEnqueued = pathMap.get(edge.dest()).timesEnqueued;
					pathMap.put(edge.dest(), newPath);
					if(!vertexQueue.contains(edge.dest())) {
						vertexQueue.offer(edge.dest());
						newPath.timesEnqueued++;
						if(newPath.timesEnqueued > graph.vertices().size() - 1) {
							// If a vertex is enqueued more than V - 1 times,
							// there is a negative cost cycle.
							throw new IllegalArgumentException(
								"Bellman-Ford cannot handle negative" +
								" cost cycles.");
						}
					}
				}
			}
		}
		if(pathMap.get(finish).cost == INFINITY) {
			// There is no path from start to finish
			return null;
		}
		// Follow the trail backwards from the Path whose destination is finish, and rebuild
		// the list of edges that constitutes the shortest path.
		LinkedList<E> shortestPath = new LinkedList<E>();
		for(V dest = finish; dest != start; ) {
			V src = pathMap.get(dest).previous;
			E edge = graph.connected(src, dest);
			shortestPath.addFirst(edge);
			dest = src;
		}
		return shortestPath;
	}
}
