package edu.cmu.cs211.graph;
import static edu.cmu.cs211.graph.GraphSearchUtil.*;
import java.util.*;

/**
 * An implementation of graph search based on Dijkstra. Does not handle negative edges.
 */
public class Dijkstra implements GraphSearch {
	/**
	 * @throws IllegalArgumentException If <tt>graph</tt> has negative edge weights.
	 */
	public <V, E extends Edge<V>> List<E> shortestPath(Graph<V, E> graph, V start, V finish) {
		// Represents a path from the start to some vertex (via another vertex), and the
		// associated cost.
		class Path implements Comparable<Path> {
			// Whether the cost of this path is known to be the least possible
			public boolean known = false;
			public V vertex, // The destination vertex
				previous; // The vertex prior the destination in the path
			// The cost of the trip; that is, the sum of the weights of the edges
			// that make up the path.
			public int cost; 
			public Path(V vertex) {
				this.vertex = vertex;
				previous = null;
				cost = INFINITY;
			}
			public Path(V vertex, int cost, V previous) {
				this.vertex = vertex;
				this.cost = cost;
				this.previous = previous;
			}
			public int compareTo(Path other) {
				// Order paths by lowest cost
				return new Integer(cost).compareTo(other.cost);
			}
		}
		// A collection of paths going from the start to every vertex in the path, where
		// the key in the map is the destination vertex.
		Map<V, Path> pathMap = new HashMap<V, Path>();
		// We use a priority queue to keep track of the path with the least cost.
		PriorityQueue<Path> pathQueue = new PriorityQueue<Path>();
		// Initialize our two collections with the vertices in the graph
		for(V vertex : graph.vertices()) {
			Path path;
			if(vertex != start) path = new Path(vertex);
			else path = new Path(start, 0, start);
			pathMap.put(vertex, path);
			pathQueue.offer(path);
		}
		while(!pathQueue.isEmpty()) {
			// Choose the best path so far and say that it is known
			Path srcPath = pathQueue.poll();
			V src = srcPath.vertex;
			srcPath.known = true;
			// We can break out of the loop if the finish vertex becomes known
			if(src == finish) break;
			// Relax the neighbors of the vertex
			for(E edge : graph.outgoingEdges(src)) {
				if(edge.weight() < 0) {
					throw new IllegalArgumentException("Dijkstra cannot handle" +
						" negative edge weights.");
				}
				V dest = edge.dest();
				Path destPath = pathMap.get(dest);
				if(destPath.known) {
					// We don't need to re-relax paths that are already known
					continue;
				}
				// See if the path via src is cheaper than the current path
				int alternateCost = safeAdd(srcPath.cost, edge.weight());
				if(alternateCost < destPath.cost) {
					// Remove the stale path from the queue
					for(Iterator<Path> iterator = pathQueue.iterator();
					iterator.hasNext(); ) {
						if(iterator.next().vertex == edge.dest())
							iterator.remove();
					}
					// Offer the updated path to the queue
					Path alternatePath = new Path(dest, alternateCost, src);
					pathMap.put(dest, alternatePath);
					pathQueue.offer(alternatePath);
				}
			}
		}
		if(pathMap.get(finish).cost == INFINITY) {
			// There is no path from start to finish
			return null;
		}
		// Follow the trail backwards from the Path whose destination is finish, and rebuild
		// the list of edges that constitutes the shortest path.
		LinkedList<E> shortestPath = new LinkedList<E>();
		for(V dest = finish; dest != start; ) {
			V src = pathMap.get(dest).previous;
			E edge = graph.connected(src, dest);
			shortestPath.addFirst(edge);
			dest = src;
		}
		return shortestPath;
	}
}
