package edu.cmu.cs211.graph;
import java.util.*;

/**
 * A class to help implement a directed graph.
 */
public class GeneralGraph<V,E extends Edge<V>> implements Graph<V,E> {
	// Maps vertices to their outgoing edges
	private Map<V, Set<E>> adjacencyMap = new HashMap<V, Set<E>>();
	/**
	 * Creates an empty graph.
	 */
	public GeneralGraph() { }
	/**
	 * Creates a graph with pre-defined vertices.
	 * @param vertices A list of the vertices in the graph.
	 * @throws NullPointerException If vertices is null, or contains null items.
	 */
	public GeneralGraph(Collection<? extends V> vertices) {
		addVertices(vertices);
	}
	/**
	 * Creates a graph with pre-defined edges and vertices.
	 * @param vertices A list of the vertices in the graph.
	 * @param edges A list of edges for the graph.
	 * @throws IllegalArgumentException If edges contains invalid edges.
	 * @throws NullPointerException If vertices or edges is null, or either contain null items.
	 */
	public GeneralGraph(Collection<? extends V> vertices, Collection<? extends E> edges) {
		addVertices(vertices);
		addEdges(edges);
	}
	/**
	 * Copy constructor.
	 * @param graph A graph to copy
	 * @throws IllegalArgumentException If <tt>graph</tt> violates Graph invariants by
	 *	returning illegal edges in its outgoingEdge methods.
	 * @throws NullPointerException If <tt>graph</tt> is null, or <tt>graph</tt>'s methods
	 *	violate Graph invariants by returning null items in verticies or outgoingEdges.
	 */
	public GeneralGraph(Graph <V, E> graph) {
		addVertices(graph.vertices());
		for(V vertex : graph.vertices())
			addEdges(graph.outgoingEdges(vertex));
	}
	public boolean addVertex(V vertex) {
		if(vertex == null) throw new NullPointerException();
		if(adjacencyMap.containsKey(vertex)) return false;
		adjacencyMap.put(vertex, new HashSet());
		return true; // The vertex was not present in the graph already.
	}
	public boolean addVertices(Collection<? extends V> vertices) {
		if(vertices == null) throw new NullPointerException();
		boolean graphChanged = false;
		for(V vertex : vertices) {
			if(addVertex(vertex)) {
				// At least one vertex was not present in the graph already;
				// this operation has changed the graph.
				graphChanged = true;
			}
		}
		return graphChanged;
	}
	public boolean addEdge(E edge) {
		if(edge == null) throw new NullPointerException();
		if(edge.src().equals(edge.dest())) {
			throw new IllegalArgumentException("An edge cannot go from a vertex" +
				" to that same vertex.");	
		}
		Set<E> outgoingEdges = adjacencyMap.get(edge.src());
		if(outgoingEdges == null || !adjacencyMap.containsKey(edge.dest())) {
			throw new IllegalArgumentException("The provided edge refers to vertices" +
				" that are not in this graph.");
		}
		for(E outgoingEdge : outgoingEdges) {
			if(edge.dest().equals(outgoingEdge.dest())) {
				// An edge from src to dest was already present!
				return true;
			}
		}
		outgoingEdges.add(edge);
		return true;
	}
	public boolean addEdges(Collection<? extends E> edges) {
		if(edges == null) throw new NullPointerException();
		boolean graphChanged = false;
		for(E edge : edges) {
			if(addEdge(edge)) {
				// At least one edge was not present in the graph already;
				// this operation has changed the graph.
				graphChanged = true;
			}
		}
		return graphChanged;
	}
	public boolean removeEdge(V src, V dest) {
		if(src == null || dest == null) throw new NullPointerException();
		Set<E> outgoingEdges = adjacencyMap.get(src);
		if(outgoingEdges == null || !adjacencyMap.containsKey(dest))
			throw new IllegalArgumentException("A provided vertex is not in the graph.");
		// We can't use Set.remove since Edge.equals depends not only on src and dest, but
		// also on the weight, which we don't know.
		for(E edge : outgoingEdges) {
			if(edge.src().equals(src) && edge.dest().equals(dest))
				return outgoingEdges.remove(edge);
		}
		return false; // The specified edge was not in the graph.
	}
	public void clearEdges() {
		for(Set<E> outgoingEdges : adjacencyMap.values())
			outgoingEdges.clear();
	}
    	public Set<V> vertices() {
		return Collections.unmodifiableSet(adjacencyMap.keySet());
	}
	public E connected(V from, V to) {
		if(from == null || to == null) throw new NullPointerException();
		Set<E> outgoingEdges = adjacencyMap.get(from);
		if(outgoingEdges == null || !adjacencyMap.containsKey(to))
			throw new IllegalArgumentException("A provided vertex is not in the graph.");
		for(E edge : outgoingEdges) {
			assert edge.src().equals(from);
			if(edge.dest().equals(to))
				return edge;
		}
		return null;
	}
	private class NeighborsSet extends AbstractSet<V> {
		private RuntimeException unmodifiable() {
			return new UnsupportedOperationException("The collection is unmodifiable.");
		}
		private RuntimeException notImplemented() {
			return new RuntimeException("The method is not implemented.");
		}
		private Set<E> edges;
		public NeighborsSet(Set<E> edges) { this.edges = edges; }
		private class Iterator implements java.util.Iterator<V> {
			private java.util.Iterator<E> edgeIterator;
			public Iterator() { edgeIterator = edges.iterator(); }
			public boolean hasNext() { return edgeIterator.hasNext(); }
			public V next() { return edgeIterator.next().dest(); }
			public void remove() { throw unmodifiable(); }
		}
		public boolean contains(Object object) {
			if(object == null) throw new NullPointerException();
			V vertex = (V)object;
			for(E edge : edges) {
				if(edge.dest().equals(vertex))
					return true;
			}
			return false;
		}
		public boolean containsAll(Collection<?> collection) {
			if(collection == null) throw new NullPointerException();
			for(Object object : collection) {
				// The set does not contain all the elements if it does not contain one of
				// the elements.
				if(!contains(object)) return false;
			}
			return true;
		}
		public int size() { return edges.size(); }
		public Object[] toArray() { throw notImplemented(); }
		public <T> T[] toArray(T[] arrayType) { throw notImplemented(); }
		public java.util.Iterator iterator() { return new Iterator(); }
		public boolean isEmpty() { return edges.isEmpty(); }
		// Mutators that throw an UnsupportedOperationException since this set should be
		// unmodifiable.
		public boolean add(V _) { throw unmodifiable(); }
		public boolean addAll(Collection<? extends V> _) { throw unmodifiable(); }
		public void clear() { throw unmodifiable(); }
		public boolean remove(Object _) { throw unmodifiable(); }
		public boolean removeAll(Collection<?> _) { throw unmodifiable(); }
		public boolean retainAll(Collection<?> _) { throw unmodifiable(); }
	}
	public Set<V> neighbors(V vertex) {
		// Use outgoingEdges(), which throws appropriate exceptions, instead of duplicating
		// code.
		Set<E> edges = (Set<E>)outgoingEdges(vertex);
		return new NeighborsSet(edges);
	}
	public Collection<E> outgoingEdges(V vertex) {
		if(vertex == null) throw new NullPointerException();
		Set<E> outgoingEdges = adjacencyMap.get(vertex);
		if(outgoingEdges == null)
			throw new IllegalArgumentException("The provided vertex is not in the graph.");
		// NOTE: The implementation of neighbors() requires that the real type of this return
		// value is a Set<E>, not a Collection<E> -- hence the usage of
		// Collections.unmodifiableSet instead of Collections.unmodifiableCollection.
		return Collections.unmodifiableSet(outgoingEdges);
	}
}
